# This quickly cobbled together script was used to create the Skepticoin Blockchain Explorer
# https://github.com/skepticoin/explorer/blob/master/README.md

from decimal import Decimal
from datetime import datetime, timezone
import os
import immutables
from collections import namedtuple
from pathlib import Path

from .datatypes import OutputReference
from .humans import human
from .params import SASHIMI_PER_COIN


PKBalance2 = namedtuple('PKBalance2', [
    'value',
    'all_output_references',
    'unspent_output_references',
    'spent_in_transactions',
])


def show_coin(sashimi):
    return "%.08f SKEPTI" % (Decimal(sashimi) / SASHIMI_PER_COIN)


def pkb2_apply_transaction(unspent_transaction_outs, public_key_balances, transaction, is_coinbase):
    with public_key_balances.mutate() as mutable_public_key_balances:
        # for coinbase we must skip the input-removal because the input references "thin air" rather than an output.
        if not is_coinbase:
            for input in transaction.inputs:
                previously_unspent_output = unspent_transaction_outs[input.output_reference]

                public_key = previously_unspent_output.public_key
                mutable_public_key_balances[public_key] = PKBalance2(
                    value=mutable_public_key_balances[public_key].value - previously_unspent_output.value,
                    all_output_references=mutable_public_key_balances[public_key].all_output_references,
                    unspent_output_references=([to for to in
                                                mutable_public_key_balances[public_key].unspent_output_references
                                                if to != input.output_reference]),

                    # in principle a single public_key could be spent more than once in a single transaction, we
                    # could change the below into a set (or alternatively, note an input index)
                    spent_in_transactions=mutable_public_key_balances[public_key].spent_in_transactions + [transaction],
                )

        for i, output in enumerate(transaction.outputs):
            output_reference = OutputReference(transaction.hash(), i)

            if output.public_key not in mutable_public_key_balances:
                mutable_public_key_balances[output.public_key] = PKBalance2(0, [], [], [])

            mutable_public_key_balances[output.public_key] = PKBalance2(
                value=mutable_public_key_balances[output.public_key].value + output.value,

                all_output_references=(mutable_public_key_balances[output.public_key].all_output_references +
                                       [output_reference]),

                unspent_output_references=(mutable_public_key_balances[output.public_key].unspent_output_references +
                                           [output_reference]),

                spent_in_transactions=mutable_public_key_balances[output.public_key].spent_in_transactions,
            )

        return mutable_public_key_balances.finish()


def pkb2_apply_block(unspent_transaction_outs, public_key_balances, block):
    public_key_balances = pkb2_apply_transaction(
        unspent_transaction_outs, public_key_balances, block.transactions[0], is_coinbase=True)

    for transaction in block.transactions[1:]:
        public_key_balances = pkb2_apply_transaction(unspent_transaction_outs, public_key_balances, transaction, False)

    return public_key_balances


def get_unspent_transaction_outs_before_block(coinstate, block):
    if block.previous_block_hash == b'\00' * 32:
        return immutables.Map()
    return coinstate.unspent_transaction_outs_by_hash[block.previous_block_hash]


def build_pkb2_block(coinstate, block, public_key_balances_2):
    # TODO factor this away.
    unspent_transaction_outs = get_unspent_transaction_outs_before_block(coinstate, block)
    return pkb2_apply_block(unspent_transaction_outs, public_key_balances_2, block)


def build_pkb2(coinstate):
    public_key_balances_2 = immutables.Map()

    for height in range(coinstate.head().height + 1):
        block = coinstate.at_head.block_by_height[height]
        public_key_balances_2 = build_pkb2_block(coinstate, block, public_key_balances_2)

    return public_key_balances_2


def build_explorer(coinstate):
    if not os.path.exists('explorer'):
        print("Created new directory for explorer")
        os.makedirs('explorer')

    public_key_balances_2 = immutables.Map()

    for height in range(coinstate.head().height + 1):
        print("Block", height)
        block = coinstate.at_head.block_by_height[height]
        potential_message = block.transactions[0].inputs[0].signature.signature

        if all([(32 <= b < 127) or (b == 10) for b in potential_message]):
            msg = "```\n" + str(potential_message, encoding="ascii") + "\n```"
        else:
            msg = ""

        unspent_transaction_outs = get_unspent_transaction_outs_before_block(coinstate, block)
        public_key_balances_2 = build_pkb2_block(coinstate, block, public_key_balances_2)

        with open(Path("explorer") / (human(block.hash()) + '.md'), 'w') as block_f:

            block_f.write(f"""## Block {human(block.hash())}

Attribute | Value
--- | ---
Height | {block.height}
Hash | {human(block.hash())}
Timestamp | {datetime.fromtimestamp(block.timestamp, tz=timezone.utc).isoformat()}
Target | {human(block.target)}
Merke root | {human(block.merkle_root_hash)}
Nonce | {block.nonce}

{msg}

### Transactions

Hash | Amount
--- | ---
""")

            for transaction in block.transactions:
                h = human(transaction.hash())
                v = show_coin(sum(o.value for o in transaction.outputs))
                block_f.write(f"""[{h}]({h}.md) | {v} \n""")

                with open(Path("explorer") / (human(transaction.hash()) + ".md"), 'w') as transaction_f:
                    transaction_f.write(f"""## Transaction {human(transaction.hash())}

In block [{human(block.hash())}]({human(block.hash())}.md)

### Inputs

Transaction | Output Index | Value | Address
--- | --- | --- | ---
""")
                    for input in transaction.inputs:
                        output_reference = input.output_reference
                        if output_reference.hash != 32 * b'\x00':
                            output = unspent_transaction_outs[output_reference]
                            h = human(output_reference.hash)
                            v = show_coin(output.value)
                            a = "SKE" + human(output.public_key.public_key) + "PTI"

                            transaction_f.write(f"""[{h}]({h}.md) | {output_reference.index} | """
                                                f"""{v} | [{a}]({a}.md)\n""")
                        else:
                            h = human(output_reference.hash)
                            v = ""
                            a = "Thin Air"

                            transaction_f.write(f"""{h} | {output_reference.index} | """
                                                f"""{v} | {a}\n""")

                    transaction_f.write(f"""### Outputs

Value | Address
--- | ---
""")
                    for output in transaction.outputs:
                        v = show_coin(output.value)
                        a = "SKE" + human(output.public_key.public_key) + "PTI"

                        transaction_f.write(f"""{v} | [{a}]({a}.md)\n""")

    for pk, pkb2 in public_key_balances_2.items():
        v = show_coin(pkb2.value)
        address = "SKE" + human(pk.public_key) + "PTI"
        with open(Path("explorer") / (address + ".md"), 'w') as address_f:
            address_f.write(f"""## {address}

Current balance: {v}
(as of block {coinstate.head().height})

## Received in

Transaction | Output Index
--- | ---
""")
            for output_reference in pkb2.all_output_references:
                h = human(output_reference.hash)
                address_f.write(f"""[{h}]({h}.md) | {output_reference.index}\n""")

            if len(pkb2.spent_in_transactions) > 0:
                address_f.write(f"""
## Spent in

Transaction | ...
--- | ---
""")

                for transaction in pkb2.spent_in_transactions:
                    address_f.write(f"""{human(transaction.hash())} | ...\n""")

            else:
                address_f.write(f"""
## Spent in

-- not spent --
""")



from ..humans import human

from .utils import (
    open_or_init_wallet,
    save_wallet,
    configure_logging_from_args,
    DefaultArgumentParser,
)


def main():
    parser = DefaultArgumentParser()
    parser.add_argument("annotation", help="Some text to help you remember a meaning for this receive address.")
    args = parser.parse_args()
    configure_logging_from_args(args)

    wallet = open_or_init_wallet()
    public_key = wallet.get_annotated_public_key(args.annotation)
    save_wallet(wallet)

    print("SKE" + human(public_key) + "PTI")



"""
The general theme of these tests is: let's do some end-to-end tests, so we at least hit a lot of code paths and expose
the most obvious of mistakes.
"""

from time import time
import logging
import os
from pathlib import Path
from time import sleep
import socket

from skepticoin.datatypes import Block
from skepticoin.coinstate import CoinState
from skepticoin.networking.threading import NetworkingThread
from skepticoin.networking.utils import load_peers_from_list


class FakeDiskInterface:
    def save_block(self, block):
        pass

    def update_peer_db(self, remote_peer):
        pass

    def save_transaction_for_debugging(self, transaction):
        pass


def _read_chain_from_disk(max_height):
    # the requirement to run these tests from an environment that has access to the real blockchain is hard-coded (for
    # now)

    coinstate = CoinState.zero()

    for filename in sorted(os.listdir('chain')):
        height = int(filename.split("-")[0])
        if height > max_height:
            return coinstate

        block = Block.stream_deserialize(open(Path('chain') / filename, 'rb'))
        coinstate = coinstate.add_block_no_validation(block)

    return coinstate


def _try_to_connect(host, port):
    # just a blocking connect to check that a server is up and listening, then close the connection.
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.connect_ex((host, port))
    sock.close()


def test_ibd_integration(caplog):
    # just testing the basics: if we set up 2 threads, one with a coinstate, and one without, will the chain propagate?

    caplog.set_level(logging.INFO)

    coinstate = _read_chain_from_disk(5)

    thread_a = NetworkingThread(coinstate, 12412, FakeDiskInterface())
    thread_a.start()

    _try_to_connect('127.0.0.1', 12412)

    thread_b = NetworkingThread(CoinState.zero(), 12413, FakeDiskInterface())
    thread_b.local_peer.network_manager.disconnected_peers = load_peers_from_list([('127.0.0.1', 12412, "OUTGOING")])
    thread_b.start()

    try:
        start_time = time()
        while True:
            if thread_b.local_peer.chain_manager.coinstate.head().height == 5:
                break

            if time() > start_time + 5:
                print("\n".join(str(r) for r in caplog.records))
                raise Exception("IBD failed")

            sleep(0.01)

    finally:
        thread_a.stop()
        thread_a.join()

        thread_b.stop()
        thread_b.join()
